==> MeatAxe-2.3.0.tar.gz: src/zts.c <==
/* ============================= C MeatAxe ==================================
   File:        $Id: zuk.c,v 1.3 1998/04/28 17:38:49 mringe Exp $
   Comment:     Tensor spin
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static char *helptext[] = {

"SYNTAX",
"    zts [<Options>] <M> <N> <Seed> <Sub>",
"",
"OPTIONS",
"    -g <#Gens>  Set number of generators (default: 2)",
"",
"FILES",
"    <M>.1,...    i  Generators on left factor",
"    <N>.1,...    i  Generators on right factor",
"    <Seed>       i  Seed vector(s)",
"    <Sub>.1,...  o  Action on subspace",
NULL};

static proginfo_t pinfo =
   { "zts", "Tensor Spin", "$Revision: $", helptext };




typedef struct
{
    int Row;	/* 1-based */
    int Col;	/* 1-based */
    FEL Mark;
} tPivotEntry;


static char *MName;		    /* Name of left factor (M) */
static char *NName;		    /* Name of right factor (N) */
static char *SeedName;		    /* Name of seed vector file */
static char *SubName;		    /* Name of result file */
static int NGen = 2;		    /* Number of generators */
static matrix_t **GenM;		    /* Generators on M (transposed) */
static matrix_t **GenN;		    /* Generators on N */
static matrix_t **Basis = NULL;	    /* Basis of invariant subspace */
static tPivotEntry *Piv = NULL;	    /* Has alwas the same size as Basis */
static int Dim = 0;		    /* Number of M-Vectors in <Basis> */
static int Src = 0;		    /* Index of vector to map next (0-based) */
static int MaxDim = 0;		    /* Capacity of <Basis> */ 
static matrix_t *Seed;		    /* Seed vectors */



static void ReadFiles()
{
    int i;
    char fn[200];
    matrix_t *m;

    GenM = NALLOC(matrix_t *,NGen);
    GenN = NALLOC(matrix_t *,NGen);
    for (i = 0; i < NGen; ++i)
    {
	sprintf(fn,"%s.%d",MName,i+1);
	m = MatLoad(fn);
	GenM[i] = MatTr(m);
	MatFree(m);
	sprintf(fn,"%s.%d",NName,i+1);
	GenN[i] = MatLoad(fn);
    }
    Seed = MatLoad(SeedName);
}


static void VectorToMatrix(PTR vec, matrix_t *m)
{
    int row, pos;
    PTR rowptr;

    zsetlen(m->noc);
    for (pos=row=1, rowptr=m->d; row <= m->nor; ++row,zadvance(&rowptr,1))
    {
	int col;
	zmulrow(rowptr,F_ZERO);
	for (col = 1; col <= m->noc; ++col, ++pos)
	{
	    FEL f = zextract(vec,pos);
	    if (f != F_ZERO)
		zinsert(rowptr,col,f);
	}
    }
}


static int FindPivot(matrix_t *m, tPivotEntry *piv)

{
    int row;
    PTR rowptr;

    zsetlen(m->noc);
    for (row = 1, rowptr=m->d; row <= m->nor; ++row, zadvance(&rowptr,1))
    {
	int col;
	FEL f;
	col = zfindpiv(rowptr,&f);
	if (col > 0)
	{
	    piv->Row = row;
	    piv->Col = col;
	    piv->Mark = f;
	    return 0;
	}
    }
    return -1;			/* Not found, zero matrix */
}


static void Clean(matrix_t *mat, const matrix_t **basis, 
    const tPivotEntry *piv, int dim)

{
    int i;
    MTXCONTEXT(Clean);

    MTX_ASSERT(dim == 0 || 
	       (mat->noc == basis[0]->noc && mat->nor == basis[0]->nor));
    zsetlen(mat->noc);
    for (i = 0; i < dim; ++i)
    {
	PTR x;
	FEL f;
	x = mat->d;
	zadvance(&x,piv[i].Row - 1);
	f = zneg(zdiv(zextract(x,piv[i].Col),piv[i].Mark));
	MatAddMul(mat,basis[i],f);
    }
}

static void Clean2(matrix_t *mat, const matrix_t **basis, 
    const tPivotEntry *piv, int dim, PTR op)

{
    int i;
    MTXCONTEXT(Clean2);

    MTX_ASSERT(mat->noc == basis[0]->noc && mat->nor == basis[0]->nor);
    zsetlen(dim);
    zmulrow(op,F_ZERO);
    zsetlen(mat->noc);
    for (i = 0; i < dim; ++i)
    {
	PTR x;
	FEL f;
	x = mat->d;
	zadvance(&x,piv[i].Row - 1);
	f = zdiv(zextract(x,piv[i].Col),piv[i].Mark);
	MatAddMul(mat,basis[i],zneg(f));
	zinsert(op,i+1,f);
    }
}


static matrix_t *Map(matrix_t *src, int gen)

{
    matrix_t *a;
    MTXCONTEXT(Map);

    MTX_ASSERT(gen >= 0 && gen < NGen);
    a = MatDup(GenM[gen]);
    MatMul(a,src);
    MatMul(a,GenN[gen]);
    return a;
}


static void CleanAndAppend(matrix_t *mat)

{
    tPivotEntry newpiv;

    Clean(mat,(const matrix_t **)Basis,Piv,Dim);
    if (FindPivot(mat,&newpiv) == 0)
    {
	if (Dim >= MaxDim)
	{
	    MaxDim += 50;
	    Basis = NREALLOC(Basis,matrix_t *,MaxDim);
	    Piv = NREALLOC(Piv,tPivotEntry,MaxDim);
	}
	Piv[Dim] = newpiv;
	Basis[Dim] = mat;
	++Dim;
	if (Dim % 100 == 0)
	    MESSAGE(2,("Dimension=%d (%d%%)\n",Dim,Src*100/Dim));
    }
    else
	MatFree(mat);
}



static void SpinUpMatrix(matrix_t *seed)

{
    int gen = 0;
    matrix_t *newvec = seed;

    zsetlen(seed->noc);
    Src = Dim;
    while (1)
    {
        CleanAndAppend(newvec);
	if (Src >= Dim)
	    break;		/* Done! */
	newvec = Map(Basis[Src],gen);
	if (++gen >= NGen)
	{
	    gen = 0;
	    ++Src;
	}
    }
}



static void Spinup()
{
    int i;
    PTR vec;

    Basis = NULL;
    Piv = NULL;
    MaxDim = 0;
    Dim = 0;
    vec = Seed->d;
    for (i = 1; i <= Seed->nor; ++i)
    {
	matrix_t *seed;
	MESSAGE(1,("Spinning up seed vector %d\n",i));
	seed = MatAlloc(zfl,GenM[0]->nor,GenN[0]->nor);
	VectorToMatrix(vec,seed);
	SpinUpMatrix(seed);		    /* <Spinup()> eats <seed>! */
	zsetlen(seed->noc);
	zadvance(&vec,1);
	if (i < Seed->nor)
	    MESSAGE(1,("Dimension = %d\n",Dim));
    }
    MESSAGE(0,("Subspace has dimension %d\n",Dim));
}



static matrix_t *CalculateAction1(int gen)
{
    matrix_t *result;
    PTR rowptr;
    int i;

    result = MatAlloc(Seed->fl,Dim,Dim);
    rowptr = result->d;
    for (i = 0; i < Dim; ++i)
    {
	matrix_t *image = Map(Basis[i],gen);
	Clean2(image,(const matrix_t **)Basis,Piv,Dim,rowptr);
	zsetlen(result->noc);
	zadvance(&rowptr,1);
    }
    return result;
}




static void CalculateAction()

{
    int i;

    MESSAGE(0,("Calculating action of generators on subspace\n"));
    for (i = 0; i < NGen; ++i)
    {
	char fn[200];
	matrix_t *result = CalculateAction1(i);
	sprintf(fn,"%s.%d",SubName,i+1);
	MESSAGE(1,("Writing %s\n",fn));
	MatSave(result,fn);
	MatFree(result);
    }
}



static void Init(int argc, char *argv[])

{
    int k;

    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((k = MtxGetOpt("g:")) != OPT_END)
    {   switch (k)
	{ 
	    case 'g':
		NGen = (int) MtxGetInt();
		break;
	}
    }

    /* Set file names
       -------------- */
    if (opt_ind + 4 != argc)
	MtxErrorExit("%s: %E",pinfo.name,ERR_USAGE);
    MName = argv[opt_ind];
    NName = argv[opt_ind+1];
    SeedName = argv[opt_ind+2];
    SubName = argv[opt_ind+3];
}

static void Cleanup()
{
}


int main(int argc, char *argv[])

{
    Init(argc, argv);
    ReadFiles();
    Spinup();
    CalculateAction();
    Cleanup();
    return 0;
}

==> MeatAxe-2.3.1.tar.gz: src/zts.c <==
/* ============================= C MeatAxe ==================================
   File:        $Id: zuk.c,v 1.3 1998/04/28 17:38:49 mringe Exp $
   Comment:     Tensor spin
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

static char *helptext[] = {

"SYNTAX",
"    zts [<Options>] <M> <N> <Seed> <Sub>",
"",
"OPTIONS",
"    -g <#Gens>  Set number of generators (default: 2)",
"",
"FILES",
"    <M>.1,...    i  Generators on left factor",
"    <N>.1,...    i  Generators on right factor",
"    <Seed>       i  Seed vector(s)",
"    <Sub>.1,...  o  Action on subspace",
NULL};

static proginfo_t pinfo =
   { "zts", "Tensor Spin", "$Revision: $", helptext };




typedef struct
{
    int Row;	/* 1-based */
    int Col;	/* 1-based */
    FEL Mark;
} tPivotEntry;


static char *MName;		    /* Name of left factor (M) */
static char *NName;		    /* Name of right factor (N) */
static char *SeedName;		    /* Name of seed vector file */
static char *SubName;		    /* Name of result file */
static int NGen = 2;		    /* Number of generators */
static matrix_t **GenM;		    /* Generators on M (transposed) */
static matrix_t **GenN;		    /* Generators on N */
static matrix_t **Basis = NULL;	    /* Basis of invariant subspace */
static tPivotEntry *Piv = NULL;	    /* Has alwas the same size as Basis */
static int Dim = 0;		    /* Number of M-Vectors in <Basis> */
static int Src = 0;		    /* Index of vector to map next (0-based) */
static int MaxDim = 0;		    /* Capacity of <Basis> */ 
static matrix_t *Seed;		    /* Seed vectors */



static void ReadFiles()
{
    int i;
    char fn[200];
    matrix_t *m;

    GenM = NALLOC(matrix_t *,NGen);
    GenN = NALLOC(matrix_t *,NGen);
    for (i = 0; i < NGen; ++i)
    {
	sprintf(fn,"%s.%d",MName,i+1);
	m = MatLoad(fn);
	GenM[i] = MatTr(m);
	MatFree(m);
	sprintf(fn,"%s.%d",NName,i+1);
	GenN[i] = MatLoad(fn);
    }
    Seed = MatLoad(SeedName);
}


static void VectorToMatrix(PTR vec, matrix_t *m)
{
    int row, pos;
    PTR rowptr;

    zsetlen(m->noc);
    for (pos=row=1, rowptr=m->d; row <= m->nor; ++row,zadvance(&rowptr,1))
    {
	int col;
	zmulrow(rowptr,F_ZERO);
	for (col = 1; col <= m->noc; ++col, ++pos)
	{
	    FEL f = zextract(vec,pos);
	    if (f != F_ZERO)
		zinsert(rowptr,col,f);
	}
    }
}


static int FindPivot(matrix_t *m, tPivotEntry *piv)

{
    int row;
    PTR rowptr;

    zsetlen(m->noc);
    for (row = 1, rowptr=m->d; row <= m->nor; ++row, zadvance(&rowptr,1))
    {
	int col;
	FEL f;
	col = zfindpiv(rowptr,&f);
	if (col > 0)
	{
	    piv->Row = row;
	    piv->Col = col;
	    piv->Mark = f;
	    return 0;
	}
    }
    return -1;			/* Not found, zero matrix */
}


static void Clean(matrix_t *mat, const matrix_t **basis, 
    const tPivotEntry *piv, int dim)

{
    int i;
    MTXCONTEXT(Clean);

    MTX_VERIFY(dim == 0 || 
	       (mat->noc == basis[0]->noc && mat->nor == basis[0]->nor));
    zsetlen(mat->noc);
    for (i = 0; i < dim; ++i)
    {
	PTR x;
	FEL f;
	x = mat->d;
	zadvance(&x,piv[i].Row - 1);
	f = zneg(zdiv(zextract(x,piv[i].Col),piv[i].Mark));
	MatAddMul(mat,basis[i],f);
    }
}

static void Clean2(matrix_t *mat, const matrix_t **basis, 
    const tPivotEntry *piv, int dim, PTR op)

{
    int i;
    MTXCONTEXT(Clean2);

    MTX_VERIFY(mat->noc == basis[0]->noc && mat->nor == basis[0]->nor);
    zsetlen(dim);
    zmulrow(op,F_ZERO);
    zsetlen(mat->noc);
    for (i = 0; i < dim; ++i)
    {
	PTR x;
	FEL f;
	x = mat->d;
	zadvance(&x,piv[i].Row - 1);
	f = zdiv(zextract(x,piv[i].Col),piv[i].Mark);
	MatAddMul(mat,basis[i],zneg(f));
	zinsert(op,i+1,f);
    }
}


static matrix_t *Map(matrix_t *src, int gen)

{
    matrix_t *a;
    MTXCONTEXT(Map);

    MTX_VERIFY(gen >= 0 && gen < NGen);
    a = MatDup(GenM[gen]);
    MatMul(a,src);
    MatMul(a,GenN[gen]);
    return a;
}


static void CleanAndAppend(matrix_t *mat)

{
    tPivotEntry newpiv;

    Clean(mat,(const matrix_t **)Basis,Piv,Dim);
    if (FindPivot(mat,&newpiv) == 0)
    {
	if (Dim >= MaxDim)
	{
	    MaxDim += 50;
	    Basis = NREALLOC(Basis,matrix_t *,MaxDim);
	    Piv = NREALLOC(Piv,tPivotEntry,MaxDim);
	}
	Piv[Dim] = newpiv;
	Basis[Dim] = mat;
	++Dim;
	if (Dim % 100 == 0)
	    MESSAGE(2,("Dimension=%d (%d%%)\n",Dim,Src*100/Dim));
    }
    else
	MatFree(mat);
}



static void SpinUpMatrix(matrix_t *seed)

{
    int gen = 0;
    matrix_t *newvec = seed;

    zsetlen(seed->noc);
    Src = Dim;
    while (1)
    {
        CleanAndAppend(newvec);
	if (Src >= Dim)
	    break;		/* Done! */
	newvec = Map(Basis[Src],gen);
	if (++gen >= NGen)
	{
	    gen = 0;
	    ++Src;
	}
    }
}



static void Spinup()
{
    int i;
    PTR vec;

    Basis = NULL;
    Piv = NULL;
    MaxDim = 0;
    Dim = 0;
    vec = Seed->d;
    for (i = 1; i <= Seed->nor; ++i)
    {
	matrix_t *seed;
	MESSAGE(1,("Spinning up seed vector %d\n",i));
	seed = MatAlloc(zfl,GenM[0]->nor,GenN[0]->nor);
	VectorToMatrix(vec,seed);
	SpinUpMatrix(seed);		    /* <Spinup()> eats <seed>! */
	zsetlen(seed->noc);
	zadvance(&vec,1);
	if (i < Seed->nor)
	    MESSAGE(1,("Dimension = %d\n",Dim));
    }
    MESSAGE(0,("Subspace has dimension %d\n",Dim));
}



static matrix_t *CalculateAction1(int gen)
{
    matrix_t *result;
    PTR rowptr;
    int i;

    result = MatAlloc(Seed->fl,Dim,Dim);
    rowptr = result->d;
    for (i = 0; i < Dim; ++i)
    {
	matrix_t *image = Map(Basis[i],gen);
	Clean2(image,(const matrix_t **)Basis,Piv,Dim,rowptr);
	zsetlen(result->noc);
	zadvance(&rowptr,1);
    }
    return result;
}




static void CalculateAction()

{
    int i;

    MESSAGE(0,("Calculating action of generators on subspace\n"));
    for (i = 0; i < NGen; ++i)
    {
	char fn[200];
	matrix_t *result = CalculateAction1(i);
	sprintf(fn,"%s.%d",SubName,i+1);
	MESSAGE(1,("Writing %s\n",fn));
	MatSave(result,fn);
	MatFree(result);
    }
}



static void Init(int argc, char *argv[])

{
    int k;

    MtxInit();
    MtxSetCommandLine(argc, argv, &pinfo);
    while ((k = MtxGetOpt("g:")) != OPT_END)
    {   switch (k)
	{ 
	    case 'g':
		NGen = (int) MtxGetInt();
		break;
	}
    }

    /* Set file names
       -------------- */
    if (opt_ind + 4 != argc)
	MtxErrorExit("%s: %E",pinfo.name,ERR_USAGE);
    MName = argv[opt_ind];
    NName = argv[opt_ind+1];
    SeedName = argv[opt_ind+2];
    SubName = argv[opt_ind+3];
}

static void Cleanup()
{
}


int main(int argc, char *argv[])

{
    Init(argc, argv);
    ReadFiles();
    Spinup();
    CalculateAction();
    Cleanup();
    return 0;
}
